/**
* local storage 存儲基本資料 & cache 資料;
* 再根據sqlite db 處理與server之同步;
* 應用於APP開發
*
*/

function HVLS(option){

	var t = $.extend({
				url:""				// web access url
				,pre:"hvls_"		// variable pre
				,debug:false		// debug
				,callback:false	// callback
				,ID: false			// 指定 ID, 網址傳遞此 ID (字串)
				,ENTITY: false 	// 指定 ENTITY，網址傳遞此ENTITY (數值)
				,userid:""			// 如果需要登入，登入帳號
				,passwd:""			// 如果需要登入，密碼
			}, option);

	var is_init = false;
	var is_lock = false;			// db 同步的時候，鎖定;

	// 以下資料存儲於 local storage;
	var sid="";						// session id
	var salt="";					// salt
	var ls=false;					// local storge items, object;
	var caches=false;				// caches, object
	var words=false;				// words
	var tables_tick={};			// tick of tables, object;
	var sync_tick=0;				// 非即時的同步

	// db 設定為 global variable

	var lock_timer = 0;			// db lock identify;
	var lock_interval_reload = false;
	var lock_interval_sync = false;

	init();


// init the items
function init(){
	// 從 local host 讀取基本設定;
	try{
		if (typeof(localStorage) == "undefined"){
			// not support HTML5;
			return is_init = -1;
		}
	} catch(err){};

	// 這些資料存儲在local storage中;
	var ar=["sid","salt","ls","caches","tables_tick","sync_tick","words"];
	for(i=0;i < ar.length; i++){
		getItem(ar[i]);
	}

	if(!sid || !ls) reload();	// 初始化立刻載入;

	if(! db || ! db.version) db_open();			// 開啟資料庫;

	// 延時 1 秒，設定循環;
	setTimeout(_set_interval, 1000);
	return true;
};

function _set_interval(){

	if(! ls){	// 第一次操作, 延後
		setTimeout(_set_interval, 60000);
		return false;
	}

	var i=parseInt(ls["interval_cache"]);
	// -1 不需要同步
	if(i >= 0){
		if(!i || i < 1) i = 1;		// min 1 minites
		i = i * 60 * 1000;
		clearInterval(lock_interval_reload);
		lock_interval_reload = setInterval(reload, i);
		_debug("interval reload "+lock_interval_reload + " every " +i+ " ms");
	}

	var i=parseInt(ls["interval_sync"]);
	// -1 不需要同步
	if(i >=0){
		if(!i || i < 1) i = 1;		// min 1 minutes;
		i = i * 60 * 1000;
		clearInterval(lock_interval_sync);
		lock_interval_sync = setInterval(sync, i);
		_debug("interval sync "+lock_interval_sync+" every "+i+" ms");
	}
};

/* 連到網路重新載入;
 * 當 tick_cache 變化的時候，載入新的cache 內容;
 * 當資料庫版本不同時，升級;
 */
function reload(){

	if(is_init <0) return false;

	var url = _getUrl("dbinit");
	_debug("reload : "+url);
	$.getJSON(url,function(data){
		if(! data.stat) return false;

		// 新的資料格式
		data=data.Data;

		// 這兩個欄位獨立處理;
		if(data.sid){
			setItem("sid",data.sid);
			setItem("salt",data.salt);
		}

		if(data.is_auth && ! sid){
			// 是否需要登入
			var url = _getUrl("dblogin");
			url=url+"&userid="+t.userid+"&passwd="+t.passwd;
			_debug("reload : "+url);
			$.getJSON(url,function(d){
				d=d.Data;
				if(d.sid){
					setItem("sid",d.sid);
					setItem("salt",d.salt);
					_doReload(data);
				}
			});
		}
		else{
			_doReload(data);
		}
	});
};

// 將 reload 移入這裡
function _doReload(data){
	if(! sid) return false;

	// 有新的cache;
	if(data.tick_cache && ( ! ls["tick_cache"] || data.tick_cache > ls["tick_cache"]) ){
		var url = _getUrl("dbcache");
		_debug("load cache : "+url);
		$.getJSON(url, function (d){
			if(! d.stat) return false;
			setItem("caches",d.Data);
		});
	}

	// 有新的 word
	if(data.word_version && (! ls["word_version"] || data.word_version > ls["word_version"])){
		var url = _getUrl("dbword");
		_debug("load word : "+url);
		$.getJSON(url, function (d){
			if(! d.stat) return false;
			setItem("words",d.Data);
		});
	}

	// 其他資料存到 ls 中;
	ls=data;	// 先設定，db_open異步操作;
	setItem("ls",data);

	db_open(1);		// db 重連，如果版本不同將會做升級操作;
	return;

};

// 從 local storage 取內容;
function getItem(a){
	if(is_init < 0 ) return false;
	// sid & salt 特別處理
	var n = _getItemName(a);
	v = $.trim(localStorage.getItem(n));
	if(! v) return false;
	var v1=v.substring(0,1);
	if(v1 == "{" || v1=="[") v=j2o(v);		// json;
	try{
		eval(a+"=v");
		return v;
	} catch(e){return v;};
};

function _getItemName(a){
	if(a == "sid" || a == "salt") return "HVAPPLS_" + a;		// sid & salt 特殊處理
	var n = t.pre;
	if(t.ID) n = n+t.ID;
	if(t.ENTITY) n = n + "_"+t.ENTITY.toString(32);
	return n + a;
};
// 設定內容到 local storage
function setItem(a, b){
	if(is_init < 0 ) return false;

	_debug("set item "+a);
	var n = _getItemName(a);
	if(typeof(b) == "object"){
		var c = JSON.stringify(b);
		localStorage.setItem(n, c);
	}
	else{
		localStorage.setItem(n, b);
	}
	try{
		eval(a+"=b");
	}catch(e){};
};

// json to object
function j2o(p_json){
	var ret={};
	try{
		return $.parseJSON(p_json);
	}catch(e){return ret;}
};

// get the url with check code;
function _getUrl(p_op){
	var r=parseInt( Math.random() * 1000);
	var url=t.url+"&op="+p_op+"&r="+r;
	if(t.ID) url = url + "&ID="+t.ID;
	if(t.ENTITY) url = url + "&ENTITY="+t.ENTITY.toString();
	if(! sid) return url+"&jsoncallback=?";
	var c=MD5(p_op+"HV,INC."+salt);
	return url+"&sid="+sid+"&chk="+c.substring(1,9)+"&jsoncallback=?";
};

// 連db, 參數設定是否立刻同步資料;
function db_open(p_sync_yn){
	if(! ls) return false;
	if(! ls["dbname"] || !ls["dbversion"]){
		is_init = true;
		return false;	// 不需要 db;
	}
	if(! db){
		try{
			db = openDatabase(ls["dbname"],"",ls["dbname"], (ls["dbsize"] ? ls["dbsize"]:1) * 1024*1024);
			_debug("db init :" + ls["dbname"]+" with version:"+db.version);
		}
		catch(e){
			return _debug("failed to open db");
		}
	}

	// reload 的時候檢查版本升級;
	var ver=$.trim(db.version);
	if(ver != "" && ver >= ls["dbversion"]){		// 沒有新的版本;
		is_init = true;			// 資料庫已準備好了;
		if(p_sync_yn) sync();	// 待同步完成才完成初始化，用 lock 來判斷是否 ready
		return;
	}

	// 初始化或者版本升級
	var url=_getUrl("dbsql");
	url = url + "&dbversion="+ver;
	_debug(url);
	return $.getJSON(url,function(data){
		if(! data.stat) return false;
		data=data.Data;
		if(! data) return false;		// 返回的是 array;
		try{
			// 返回的是 array;
			max_cnt = data.length;
		}
		catch(e){
			return false;
		}

		setLock();
		// 先變更版本，再執行 sql;
		_debug("begin db version: "+ ver + " to "+ls["dbversion"]);
		return db.changeVersion(ver,ls["dbversion"], function(tx){
			// _debug("db init begin");
			// sql 語句在 data[] 中;
			var cnt = 0;
			for(k=0;k<max_cnt;k++){
				_debug(data[k]);
				tx.executeSql(data[k],[],function(t){
					cnt++;
					_debug(cnt + ":" + max_cnt);
					if(cnt == max_cnt){
						is_init = true;	// 資料庫準備好了
						unlock();			// 操作完成
						setTimeout(sync,75);	// lock 判斷同步資料的過程
						return true;
					}
				},function(){
					_debug("failed at sql execute");
				});
			}
		},function(data){
			_debug("version changed error");
		});
		// change the version
	})
};

// 讀入 db;
function sync(){
	if(! db) return false;

	if(! ls["tables"] && ! ls["tables_cache"]){
		// 沒有需要同步的 tables;
		if(is_init == false) is_init = true;
		return false;
	}

	// 採用 lock 方式來管理同步的排隊情況;
	if(ls["tables"]){
		for(i in ls["tables"]){
			var table=ls["tables"][i];
			var tick=tables_tick[table];
			if(! tick) tick = 0;
			_sync(table,tick,1);
		}
	}
	// 非即時的同步
	if(ls["tables_cache"]) _sync2(sync_tick,1);
	return;
};

// 分頁取得資料;
function _sync(p_table, p_tick, p_page){

	if( isLock()) {
		setTimeout(function(){
			_sync(p_table, p_tick, p_page);
		}, 100);	// 資料庫鎖定中，延時再試，解決排隊問題;
		return false;
	}

	setLock();
	var url = _getUrl("dbsync");
	var ver = db.version;
	url=url+"&table="+p_table+"&tick="+p_tick+"&page="+p_page+"&dbversion="+ver;
	_debug(url);

	$.getJSON(url,function(data){
		if(! data.stat) return unlock();		// no records

		// 儲存最新的tick;
		if(data.tick && tables_tick[p_table] != data.tick){
			if(! tables_tick || typeof(tables_tick) != "object" ) tables_tick = {};
			tables_tick[p_table] = data.tick;
			_debug(tables_tick[p_table]+":"+p_table+":table, tick");
			setItem("tables_tick", tables_tick);
		}

		next = data.next;	// 是否需要翻頁

		// 資料在Data
		d=data.Data;
		if(! d) return unlock();

		_do_sync(p_table, d, p_tick, p_page,next,1);
	},function(){
		unlock();
	});	// get json
};

// 分頁取得資料;
function _sync2(p_tick, p_page){

	if( isLock()) {
		setTimeout(function(){
			_sync2( p_tick, p_page);
		}, 100);	// 資料庫鎖定中，延時再試，解決排隊問題;
		return false;
	}

	setLock();
	var url = _getUrl("dbsync2");
	var ver = db.version;
	url=url+"&tick="+p_tick+"&page="+p_page+"&dbversion="+ver;
	_debug(url);

	$.getJSON(url,function(data){
		if(! data.stat) return unlock();		// no records

		// 儲存最新的tick;
		if(data.tick && sync_tick != data.tick){
			setItem("sync_tick", data.tick);
		}

		next = data.next;	// 是否需要翻頁

		// 資料在Data
		d=data.Data;
		if(! d) return unlock();

		for(table in d){
			data = d[table];
			if(! data) continue;
			_do_sync(table, data, p_tick, p_page,next,2);
			next = 0;	// 只需要一次操作
		}
	},function(){
		unlock();
	});
};

// 執行 sql;
function _do_sync(p_table, data,p_tick,p_page, p_next_yn,p_type){

	var sql_del = "delete from "+p_table+" where seq in (0";
	var sql_ins = [];

	try{
		cnts = data.length;	// 返回的是array
	}catch(e){
		return false;
	};

	for(i=0; i< cnts; i++){

		var rs = data[i];
		if(typeof(rs) != "object") continue;

		// 組合語句，然後先刪除，再新增;
		sql_del = sql_del + ","+rs["seq"];

		// 再新增;
		if(! rs["deleted"]){
			var sql=s1=s2='';
			for(k in rs){
				s1=s1+k+',';
				s2=s2+'"'+addslashes(rs[k])+'",';
			}
			sql_ins.push("INSERT INTO "+p_table+" ("+s1.substring(0,s1.length -1)+") values ("+s2.substring(0,s2.length -1)+")");
		}
	}

	db.transaction(function(tx){
		var sql = sql_del+")";
		// 先刪除，再新增;
		// _debug(sql);
		tx.executeSql(sql,[],function(t){
		//	_debug(sql_ins.length+": to be updated");
			if(! sql_ins.length) return unlock();	// 無新增的資料;
			// 等待刪除之後再新增
			for(k = 0; k < sql_ins.length; k++){
			//	_debug(sql_ins[k]);
				t.executeSql(sql_ins[k]);
			}	// end of for
			// 延時等待 insert 完成;
			var t1 = Math.min(sql_ins.length * 3,1000);		// 粗略估算延時時間
			// _debug(t1 + " later to do unlock");
			setTimeout(
				function(){
					_debug(cnts + " rows updated");
					unlock();		// 解除鎖定;

					// 翻頁
					if(p_next_yn){
						p_page ++;
						if(p_type == 2){
							_sync2(p_tick, p_page);		// 翻頁循環
						}
						else{
							_sync(p_table, p_tick, p_page);		// 翻頁循環
						}
					}
				},t1);
			},function(){
				return unlock();
		});	// end of sql_del
	});	// do transaction
};

function addslashes(p_str){
	if(! p_str) return p_str;
	return p_str.replace(/"/g, '""');
};

function isLock(){
	return is_lock;
};
function setLock(){
	_debug("set lock");
	is_lock = true;
	lock_timer = setTimeout(function(){
		is_lock = false;
	}, 60*1000);
	return;
};
function unlock(){
	_debug("unlock");
	is_lock = false;
	clearTimeout(lock_timer);
};

// debug log
function _debug(p_msg){
	if(! t.debug) return false;
	var d=new Date();
	var c=d.getHours()+":"+d.getMinutes()+":"+d.getSeconds()+"."+d.getMilliseconds();
 $("#log").append("<p>"+c+"      "+p_msg+"</p>");

	return console.log(c+"	"+p_msg);
};

// reset the db & cache; 不提供外部使用;
function reset(){

	if(is_init < 0) return false;

	// db version 改為 0
	db.changeVersion(db.version,"",function(){
		setItem("ls",{});
		setItem("caches",{});
		setItem("words",{});
		setItem("tables_tick",{});
		setItem("sync_tick",0);
		is_init = false;
		reload();
	});
};

// 對外初始化;
function setup(callback){
	t.callback = callback;
};
// 取得 object 的長度
function _get_obj_length(p_obj){
	if(typeof(p_obj) != "object") return false;
	var cnt=0;
	for(i in p_obj){
		cnt++;
	}
	return cnt;
}

// 定時判斷是否已經初始化完成。
function chkReady(p_cnt1, p_cnt2){
	_debug("chk ready called : "+ p_cnt1 + ","+p_cnt2);
	if(! t.callback) return false;
	if(! p_cnt1) p_cnt1 = 0;
	if(! p_cnt2) p_cnt2 = 0;
	if(is_init != 1){
		if(p_cnt1 > 50) return t.callback({stat:"fail"});	// 初始化只是準備db, 最多 5秒;
		p_cnt1 ++;
		setTimeout(function(){ chkReady(p_cnt1,p_cnt2) },100);	// 每隔0.1秒檢查1次;
		return false;
	}
	// 連續沒有鎖定；用於顯示歡迎畫面;
	if(is_lock){
		p_cnt2 = 0;
		setTimeout(function(){ chkReady(p_cnt1,p_cnt2) },50);	// 每隔0.05秒檢查1次;
		return false;
	}
	if(p_cnt2 < 2) {
		p_cnt2 ++ ;
		setTimeout(function(){ chkReady(p_cnt1,p_cnt2) },50);	// 連續3次沒有鎖定
		return false;
	}
	return t.callback({stat:"success"});
};

	// public methods
	return {
		isInit:function(){	// 是否初始化完成：HTML5 support + 連上網路
			return is_init;
		}
		,isLock:function(){
			return is_lock;
		}
		,reload:function(callback){		// 重新載入
			setup(callback);
			reload();
			return chkReady();
		}
		// 檢查是否準備好了
		,chkReady:function(callback){
			setup(callback);
			return chkReady();
		}
		// 以下需要考慮是否完成初始化操作;
		,cache:function(table){	// cache 的內容;
			return caches[table];
		}
		,set:function(key,val){
			return setItem(key,val);
		}
		,ls:function(item){		// local storage 內容;
			return ls[item];
		}
		,get:function(key){
			return getItem(key);
		}
		,word:function(key){
			var r = words[key];
			if(! r) return key;
			return r;
		}
	};

};		// end of function ls()

